# -*- coding: utf-8 -*-
from typing import Iterator, Any
import logging

from wood.invalidate import PrefixInvalidator


logger = logging.getLogger(__name__)


class CloudFrontInvalidator(PrefixInvalidator):
    """
    Invalidates paths in a CloudFront distribution.
    """

    def __init__(self, client: Any, distribution: str, reference: str):
        """
        Initialise a new CloudFront invalidator.

        :param client: The boto3 CloudFront client to use when interacting with
                       the service.
        :param distribution: The distribution ID to invalidate.
        :param reference: A description for any invalidations performed using
                          this instance.
        """
        self._client = client
        self._distribution = distribution
        self._reference = reference

    def _invalidate_prefixes(self, prefixes: Iterator[str]) -> None:
        prefixes = list(['/' + prefix for prefix in prefixes])
        if not prefixes:
            # nothing to do (CloudFront will complain if given an empty list)
            return
        logger.info('Invalidating %d prefixes (%s)', len(prefixes),
                    ', '.join(prefixes))
        response = self._client.create_invalidation(
            DistributionId=self._distribution,
            InvalidationBatch={
                'Paths': {
                    'Quantity': len(prefixes),
                    'Items': prefixes
                },
                'CallerReference': self._reference
            })
        logger.debug('CloudFront invalidation response: %s', response)
        logger.info('Created invalidation %s', response['Invalidation']['Id'])



# -*- coding: utf-8 -*-
from typing import List, Dict, Union, Any, Iterable, Tuple
import pathlib
import logging
import mimetypes

from wood import util
from wood.comparison import Comparison
from wood.entities import Root, Directory, File, Entity
from wood.sync import Syncer


logger = logging.getLogger(__name__)


def objects_to_root(objects: List) -> Root:
    """
    Convert a list of s3 ObjectSummaries into a directory tree.

    :param objects: The list of objects, e.g. the result of calling
                    `.objects.all()` on a bucket.
    :return: The tree structure, contained within a root node.
    """

    def _to_tree(objs: Iterable) -> Dict:
        """
        Build a tree structure from a flat list of objects.

        :param objs: The raw iterable of S3 `ObjectSummary`s, as returned by a
                     bucket listing.
        :return: The listing as a nested dictionary where keys are directory
                 and file names. The values of directories will in turn be a
                 dict. The values of keys representing files will be the
                 `ObjectSummary` instance.
        """
        path_tree = {}
        for obj in objs:
            is_dir = obj.key.endswith('/')
            chunks = [chunk for chunk in obj.key.split('/') if chunk]
            chunk_count = len(chunks)
            tmp = path_tree
            for i, chunk in enumerate(chunks):
                is_last_chunk = i == chunk_count - 1

                if is_last_chunk and not is_dir:
                    tmp[chunk] = obj
                else:
                    # must be a directory
                    if chunk not in tmp:
                        # it doesn't exist - create it
                        tmp[chunk] = {}
                    tmp = tmp[chunk]
        return path_tree

    def _to_entity(key: str, value: Union[Dict, Any]) -> Entity:
        """
        Turn a nested dictionary representing an S3 bucket into the correct
        `Entity` object.

        :param key: The name of the entity.
        :param value: If the entity is a directory, the nested dict
                      representing its contents. Otherwise, the `ObjectSummary`
                      instance representing the file.
        :return: The entity representing the entity name and value pair.
        """
        if isinstance(value, dict):
            return Directory(
                key,
                {key_: _to_entity(key_, value_)
                 for key_, value_ in value.items()})

        return File(pathlib.PurePath(value.key).name, value.size,
                    value.e_tag.strip('"'))

    tree = _to_tree(objects)
    return Root({pathlib.PurePath(key).name: _to_entity(key, value)
                 for key, value in tree.items()})


class S3Syncer(Syncer[Root, Directory]):
    """
    Synchronises a local directory with an S3 bucket.
    """

    # the maximum number of keys that can be specified in a single delete call
    _MAX_DELETES_PER_REQUEST = 1_000

    def __init__(self, base: pathlib.PurePath, bucket, prefix: str = ''):
        """
        Initialise a new S3 syncer.

        :param base: The local base directory.
        :param bucket: The bucket to upload to.
        :param prefix: The prefix within which to work, within the bucket.
                       Defaults to no prefix, i.e. the root of the bucket.
        """
        self._base = base
        self._bucket = bucket
        self._prefix = prefix

    def _delete(self, paths: Iterable[str]) -> None:
        """
        Delete a collection of paths from S3.

        :param paths: The paths to delete. The prefix will be prepended to each
                      one.
        :raises ClientError: If any request fails.
        """
        for chunk in util.chunk(paths, self._MAX_DELETES_PER_REQUEST):
            keys = list([self._prefix + key for key in chunk])
            logger.info('Deleting %d objects (%s)', len(keys), ', '.join(keys))
            response = self._bucket.delete_objects(Delete={
                'Objects': [{'Key': key} for key in keys],
                'Quiet': True
            })
            logger.debug('Delete objects response: %s', response)

    def _upload(self, items: Iterable[Tuple[str, str]]) -> None:
        """
        Upload a collection of paths to S3.

        :param items: An iterable of pairs containing the local path of the
                      file to upload, and the remote path to upload it to. The
                      prefix will be appended to each remote path.
        """
        for src, key in items:
            logger.info(f'Uploading {src} to {key}')
            mimetype, _ = mimetypes.guess_type(src)
            if mimetype is None:
                logger.warning(f'Could not guess MIME type for {src}')
                mimetype = 'application/octet-stream'

            logger.debug(f'Deduced MIME type: {mimetype}')
            self._bucket.upload_file(src, key, ExtraArgs={
                    'ContentType': mimetype
                })

    def sync(self, comparison: Comparison[Root, Directory]) -> None:
        self._delete(comparison.deleted(include_directories=False))
        self._upload(zip(comparison.new(self._base,
                                        include_intermediates=False),
                         comparison.new(include_intermediates=False)))
        self._upload(zip(comparison.modified(self._base),
                         comparison.modified()))



# -*- coding: utf-8 -*-
from typing import Generic, TypeVar, Iterator, IO, Optional, Dict
import abc
import sys
import pathlib
import itertools

import os

from wood import util
from wood.entities import Entity, File, Directory

L = TypeVar('L', bound=Entity)
R = TypeVar('R', bound=Entity)
L1 = TypeVar('L1', bound=Entity)
R1 = TypeVar('R1', bound=Entity)


class Comparison(Generic[L, R], metaclass=abc.ABCMeta):
    """
    Represents a comparison between any two entities, e.g. a File and a File,
    a File and Directory etc.
    """

    # the number of spaces to indent per level
    _INDENT_SIZE = 4

    def __init__(self, left: Optional[L], right: Optional[R]):
        """
        Initialise a new comparison.

        :param left: The left or "original"/"old" entity. Omit if the entity
                     is new.
        :param right: The right or "current"/"new" entity. Omit if the entity
                      has been deleted.
        :raises ValueError: If both the left and right entities are None.
        """
        if left is None and right is None:
            raise ValueError('The left and right side cannot both be None')
        self.left = left
        self.right = right

    @property
    def is_empty(self) -> bool:
        """
        Find whether this entity is empty.

        :return: True if the entity is empty, false otherwise.
        """
        raise NotImplementedError()

    @property
    def is_new(self) -> bool:
        """
        Find whether this comparison represents a new file.

        :return: True if the right is a new file, false otherwise.
        """
        return self.left is None

    def new(self, base: pathlib.PurePath = pathlib.PurePath(),
            include_intermediates: bool = True) -> Iterator[str]:
        """
        Find the list of new paths in this comparison.

        :param base: The base directory to prepend to the right entity's name.
        :param include_intermediates: Whether to include new non-empty
                                      directories in the returned iterable. If
                                      you only care about files, or are using
                                      flat key-based storage system like S3
                                      where directories are a made-up concept,
                                      this can be set to false.
        :return: An iterator of the new paths.
        """
        if self.is_new:
            yield str(base / self.right.name)

    @property
    @abc.abstractmethod
    def is_modified(self) -> bool:
        """
        Find whether this comparison is for a modified file. Note that
        directory comparisons are always not modified, as a directory itself
        can only be added (new) or removed (deleted).

        :return: True if this comparison represents a modified file, false
                 otherwise.
        """
        raise NotImplementedError()

    def modified(self, base: pathlib.PurePath = pathlib.PurePath()) \
            -> Iterator[str]:
        """
        Find the paths of modified files. There is no option to include
        intermediate directories, as all files and directories exist in both
        the left and right trees.

        :param base: The base directory to recursively append to the right
                     entity.
        :return: An iterable of paths of modified files.
        """
        # N.B. this method will only ever return files, as directories cannot
        # be "modified"
        if self.is_modified:
            yield str(base / self.right.name)

    @property
    def is_deleted(self) -> bool:
        """
        Find whether this comparison represents a deleted entity.

        :return: True if the left entity has been deleted, false otherwise.
        """
        return self.right is None

    def deleted(self, base: pathlib.PurePath = pathlib.PurePath(),
                include_children: bool = True,
                include_directories: bool = True) -> Iterator[str]:
        """
        Find the paths of entities deleted between the left and right entities
        in this comparison.

        :param base: The base directory to recursively append to entities.
        :param include_children: Whether to recursively include children of
                                 deleted directories. These are themselves
                                 deleted by definition, however it may be
                                 useful to the caller to list them explicitly.
        :param include_directories: Whether to include directories in the
                                    returned iterable.
        :return: An iterable of deleted paths.
        """
        if self.is_deleted:
            yield str(base / self.left.name)

    @property
    @abc.abstractmethod
    def invalidate(self) -> bool:
        """
        Find whether this entity should be invalidated.

        :return: True if it should be, false otherwise.
        """
        raise NotImplementedError()

    def invalidations(self) -> Iterator[str]:
        """
        Get a set of invalidation prefixes for this entity and its children.
        These prefixes will resemble relative paths - they will not have a
        leading slash.

        :return: The set of invalidations.
        """
        if self.invalidate:
            yield self.left.name

    @staticmethod
    def compare(left: Optional[L], right: Optional[R]) -> 'Comparison[L, R]':
        """
        Calculate the comparison of two entities.

        | left      | right     | Return Type             |
        |===========|===========|=========================|
        | file      | file      | FileComparison          |
        | file      | directory | FileDirectoryComparison |
        | file      | None      | FileComparison          |
        | directory | file      | DirectoryFileComparison |
        | directory | directory | DirectoryComparison     |
        | directory | None      | DirectoryComparison     |
        | None      | file      | FileComparison          |
        | None      | directory | DirectoryComparison     |
        | None      | None      | TypeError               |

        :param left: The left side or "before" entity.
        :param right: The right side or "after" entity.
        :return: See table above.
        """
        if isinstance(left, File) and isinstance(right, Directory):
            return FileDirectoryComparison(left, right)

        if isinstance(left, Directory) and isinstance(right, File):
            return DirectoryFileComparison(left, right)

        if isinstance(left, File) or isinstance(right, File):
            return FileComparison(left, right)

        if isinstance(left, Directory) or isinstance(right, Directory):
            return DirectoryComparison(left, right)

        raise TypeError(f'Cannot compare entities: {left}, {right}')

    def print_hierarchy(self, level: int = 0, file: IO[str] = sys.stdout) \
            -> None:
        """
        Print this comparison and its children with indentation to represent
        nesting.

        :param level: The level of indentation to use. This is mostly for
                      internal use, but you can use it to inset the root
                      comparison.
        :param file: The stream to print to. Defaults to stdout.
        """
        print(' ' * self._INDENT_SIZE * level + str(self), file=file)

    def __str__(self):
        return f'{self.__class__.__name__}(' \
               f'{self.left or None}|{self.right or None}, ' \
               f'is_new: {self.is_new}, ' \
               f'is_modified: {self.is_modified}, ' \
               f'is_deleted: {self.is_deleted}, ' \
               f'invalidate: {self.invalidate})'


class FileComparison(Comparison[File, File]):
    """
    A comparison of two files.
    """

    @property
    def is_empty(self) -> bool:
        # we have no insight into the contents of files
        return False

    @property
    def is_modified(self) -> bool:
        """
        Find whether the files on the left and right are different. Note,
        modified implies the contents of the file have changed, which is
        predicated on the file existing on both the left and right. Therefore
        this will be false if the file on the left has been deleted, or the
        file on the right is new.

        :return: Whether the file has been modified.
        """
        if self.is_new or self.is_deleted:
            return False
        return self.left.md5 != self.right.md5

    @property
    def invalidate(self) -> bool:
        return self.is_modified or self.is_deleted

    def __init__(self, left: Optional[File], right: Optional[File]):
        """
        Initialise a new file comparison.

        :param left: The "original" or "old" file. Cannot be None if right is
                     None.
        :param right: The "current" or "new" file. Cannot be None if left is
                      None.
        """
        super().__init__(left, right)


class DirectoryComparison(Comparison[Directory, Directory]):
    """
    A comparison of two directories.
    """

    @property
    def is_empty(self) -> bool:
        return not self.children

    def new(self, base: pathlib.PurePath = pathlib.PurePath(),
            include_intermediates: bool = True) -> Iterator[str]:

        # if the RHS has been deleted, there's nothing to do, so yield nothing
        if self.is_deleted:
            return

        us = base / self.right.name

        # if we ourselves are new, we need to return our path, as long as we
        # don't have any children, unless include_intermediates is
        # true
        if self.is_new and (not self.children or include_intermediates):
            # we have no children, or we do but include_intermediates is true
            # N.B. we can guarantee intermediate directories will be seen
            # before their children
            yield str(us) + '/'

        # we need to return all new paths in our children
        # N.B. if we are new, all of our children have to be new by definition
        yield from itertools.chain.from_iterable(
            [child.new(us, include_intermediates)
             for child in self.children.values()])

    @property
    def is_modified(self) -> bool:
        # we regard directories as containers - they can be new or deleted, but
        # never modified
        return False

    def modified(self, base: pathlib.PurePath = pathlib.PurePath()) \
            -> Iterator[str]:
        # given we cannot be modified, provided we are not deleted, we simply
        # delegate to our children
        if not self.is_deleted:
            us = base / self.right.name
            yield from itertools.chain.from_iterable(
                [child.modified(us) for child in self.children.values()])

    def deleted(self, base: pathlib.PurePath = pathlib.PurePath(),
                include_children: bool = True,
                include_directories: bool = True) -> Iterator[str]:
        if self.is_new:
            return  # no self.left

        us = base / self.left.name

        if self.is_deleted and include_directories:
            yield str(us) + '/'

        if not self.is_deleted or include_children:
            yield from itertools.chain.from_iterable(
                [child.deleted(us, include_children, include_directories)
                 for child in self.children.values()])

    def __init__(self, left: Optional[Directory], right: Optional[Directory]):
        super().__init__(left, right)
        left_children = {} if left is None else left.children
        right_children = {} if right is None else right.children
        self.children = {
            name: Comparison.compare(l, r)
            for name, (l, r) in
            util.zip_dict(
                left_children,
                right_children).items()}  # type: Dict[str, Comparison[Entity]]

    def print_hierarchy(self, level: int = 0, file: IO[str] = sys.stdout) \
            -> None:
        super().print_hierarchy(level, file)
        for child in self.children.values():
            child.print_hierarchy(level + 1, file)

    @property
    def invalidate(self) -> bool:
        # N.B. the case of no children is interesting:
        #  - if we allow ourselves to be invalidated, this allows the
        #    possibility of an aggregation on our parent directory
        #  - however, if that aggregation could not be done, we've incurred an
        #    explicit invalidation ($0.005) for ourselves for no reason
        # So it's not cut and dry. This method goes for the first case, and we
        # account for the second later.
        return all([child.invalidate for child in self.children.values()])

    def invalidations(self) -> Iterator[str]:
        name = self.right.name if self.is_new else self.left.name
        if self.invalidate:
            # could also just `yield name + '*'`, however this could
            # over-invalidate if multiple directories have a prefix of `name`

            # hack: the root directory is an empty string; we shouldn't have to
            # worry about this here
            if name:
                yield name

            yield os.path.join(name, '*')
        else:
            child_invalidations = itertools.chain.from_iterable(
                [child.invalidations()
                 for child in self.children.values()
                 if not child.is_empty])  # exclude unchanged empty directories
            yield from [os.path.join(name, invalidation)
                        for invalidation in child_invalidations]


class ReplacementComparison(Comparison[L, R], Generic[L, R],
                            metaclass=abc.ABCMeta):
    """
    A comparison between two entities of differing types. Due to the
    incompatibility, we simply treat this as the left entity being deleted,
    and the right entity being new.
    """

    @property
    def is_new(self) -> bool:
        return True

    @property
    def is_modified(self) -> bool:
        return False

    @property
    def is_deleted(self) -> bool:
        return True

    @property
    def invalidate(self) -> bool:
        return True

    def __init__(self, left: L, right: R,
                 children: Dict[str, Comparison[L1, R1]]):
        super().__init__(left, right)
        self.children = children

    def print_hierarchy(self, level: int = 0, file: IO[str] = sys.stdout) \
            -> None:
        super().print_hierarchy(level, file)
        for child in self.children.values():
            child.print_hierarchy(level + 1, file)


class FileDirectoryComparison(ReplacementComparison[File, Directory]):
    """
    The comparison of a file and directory.
    """

    @property
    def is_empty(self) -> bool:
        return not self.right.children

    def new(self, base: pathlib.PurePath = pathlib.PurePath(),
            include_intermediates: bool = True) -> Iterator[str]:
        us = base / self.right.name

        # if we ourselves are new, we need to return our path, as long as we
        # don't have any children, unless include_intermediates is
        # true
        if not self.children or include_intermediates:
            # we have no children, or we do but include_intermediates is true
            # N.B. we can guarantee intermediate directories will be seen
            # before their children
            yield str(us) + '/'

        # we need to return all new paths in our children
        # N.B. if we are new, all of our children have to be new by definition
        yield from itertools.chain.from_iterable(
            [child.new(us, include_intermediates)
             for child in self.children.values()])

    def __init__(self, left: File, right: Directory):
        super().__init__(left, right,
                         {name: Comparison.compare(None, r)
                          for name, r in right.children.items()})


class DirectoryFileComparison(ReplacementComparison[Directory, File]):
    """
    The comparison of a directory and file.
    """

    @property
    def is_empty(self) -> bool:
        return False

    def deleted(self, base: pathlib.PurePath = pathlib.PurePath(),
                include_children: bool = True,
                include_directories: bool = True) -> Iterator[str]:
        us = base / self.left.name

        if self.is_deleted and include_directories:
            yield str(us) + '/'

        if not self.is_deleted or include_children:
            yield from itertools.chain.from_iterable(
                [child.deleted(us, include_children, include_directories)
                 for child in self.children.values()])

    def invalidations(self) -> Iterator[str]:
        if self.invalidate:
            yield self.left.name + '/*'

    def __init__(self, left: Directory, right: File):
        super().__init__(left, right,
                         {name: Comparison.compare(l, None)
                          for name, l in left.children.items()})



# -*- coding: utf-8 -*-
from typing import Iterator
import abc

from wood.comparison import Comparison
from wood.entities import Entity


class Invalidator(metaclass=abc.ABCMeta):
    """
    Implemented by things that know how to invalidate the changes in a
    comparison.
    """

    def invalidate(self, comparison: Comparison[Entity, Entity]) -> None:
        """
        Invalidate as necessitated by a comparison.

        :param comparison: The comparison whose changes to invalidate.
        """
        raise NotImplementedError()


class PrefixInvalidator(Invalidator, metaclass=abc.ABCMeta):
    """
    Implemented by things that can invalidate a regex-like prefix suffixed with
    an asterisk. This allows a more efficient (and cheaper) invalidation by
    grouping paths together where possible.
    """

    def invalidate(self, comparison: Comparison[Entity, Entity]) -> None:
        self._invalidate_prefixes(comparison.invalidations())

    @abc.abstractmethod
    def _invalidate_prefixes(self, prefixes: Iterator[str]) -> None:
        """
        Purge a collection of prefixes.

        :param prefixes: The collection of prefixes to purge. These will
                         resemble relative paths that may or may not end with
                         an asterisk, meaning a wildcard.
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from typing import Dict, Tuple, Optional, TypeVar, Iterable

import itertools
import more_itertools

A = TypeVar('A')
B = TypeVar('B')


def zip_dict(a: Dict[str, A], b: Dict[str, B]) \
        -> Dict[str, Tuple[Optional[A], Optional[B]]]:
    """
    Combine the values within two dictionaries by key.

    :param a: The first dictionary.
    :param b: The second dictionary.
    :return: A dictionary containing all keys that appear in the union of a and
             b. Values are pairs where the first part is a's value for the key,
             and right second part b's value.
    """
    return {key: (a.get(key), b.get(key)) for key in a.keys() | b.keys()}


def chunk(iterable: Iterable[A], n: int) \
        -> Iterable[more_itertools.more.peekable]:
    """
    Produce an iterable of interables of a maximum length from a (presumably
    longer) iterable. This is useful when only so many elements can be
    processed at once, such as an API that limits to n things per request.

    :param iterable: The iterable to chunk into iterables of size up to n.
    :param n: The maximum length of each iterable.
    :return: An iterable of iterables. Each iterable will be of size n, except
             possibly the last one which will contain fewer elements.
    """
    iterator = iter(iterable)
    while True:
        chunk_ = more_itertools.peekable(itertools.islice(iterator, n))
        try:
            chunk_.peek()
        except StopIteration:
            return
        yield chunk_



# -*- coding: utf-8 -*-
import abc
import hashlib
import itertools
import pathlib
import sys
from typing import Optional, IO, Iterator, Dict


class Entity(metaclass=abc.ABCMeta):
    """
    Represents a snapshot of a file or directory.
    """

    # the number of spaces to indent per level
    _INDENT_SIZE = 4

    def __init__(self, name):
        """
        Initialise a new entity.

        :param name: The name of the entity, e.g. "essay.txt" or "photos".
        """
        self.name = name

    @abc.abstractmethod
    def walk_paths(self,
                   base: Optional[pathlib.PurePath] = pathlib.PurePath()) \
            -> Iterator[pathlib.PurePath]:
        """
        Recursively traverse all paths inside this entity, including the entity
        itself.

        :param base: The base path to prepend to the entity name.
        :return: An iterator of paths.
        """
        raise NotImplementedError()

    def _walk_paths(self, base: pathlib.PurePath) \
            -> Iterator[pathlib.PurePath]:
        """
        Internal helper for walking paths. This is required to exclude the name
        of the root entity from the walk.

        :param base: The base path to prepend to the entity name.
        :return: An iterator of paths.
        """
        return self.walk_paths(base)

    @abc.abstractmethod
    def walk_files(self) -> Iterator['File']:
        """
        Recursively traverse all files inside this entity, including the entity
        itself.

        :return: An iterator of files.
        """
        raise NotImplementedError()

    @classmethod
    def from_path(cls, path: pathlib.Path) -> 'Entity':
        """
        Create an entity from a local path.

        :param path: The path to the entity, either a file or directory.
        :return: An entity instance representing the path.
        """
        if path.is_file():
            return File.from_path(path)
        return Directory.from_path(path)

    def print_hierarchy(self, level: int = 0, file: IO[str] = sys.stdout) \
            -> None:
        """
        Print this entity and its children with indentation to represent
        nesting.

        :param level: The level of indentation to use. This is mostly for
                      internal use, but you can use it to inset the root
                      entity.
        :param file: The stream to print to. Defaults to stdout.
        """
        print(' ' * self._INDENT_SIZE * level + str(self), file=file)

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.name})'

    def __repr__(self) -> str:
        return f'<{self}>'


class File(Entity):
    """
    Represents a file.
    """

    def __init__(self, name, size, md5):
        """
        Initialise a new file entity.

        :param name: The file name, including extension if any.
        :param size: The size of the file in bytes.
        :param md5: The lowercase hex representation of the file's MD5
                    checksum. Should be exactly 32 chars long.
        """
        super().__init__(name)
        self.size = size
        self.md5 = md5

    def walk_paths(self,
                   base: Optional[pathlib.PurePath] = pathlib.PurePath()) \
            -> Iterator[pathlib.PurePath]:
        yield base / self.name

    def walk_files(self) -> Iterator['File']:
        yield self

    @staticmethod
    def _md5(path: pathlib.PurePath):
        """
        Calculate the MD5 checksum of a file.

        :param path: The path of the file whose checksum to calculate.
        :return: The lowercase hex representation of the file's MD5
                    checksum, exactly 32 chars long.
        """
        hash_ = hashlib.md5()
        with open(path, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b''):
                hash_.update(chunk)
        return hash_.hexdigest()

    @classmethod
    def from_path(cls, path: pathlib.Path) -> 'File':
        """
        Create a file entity from a file path.

        :param path: The path of the file.
        :return: A file entity instance representing the file.
        :raises ValueError: If the path does not point to a file.
        """
        if not path.is_file():
            raise ValueError('Path does not point to a file')
        return File(path.name, path.stat().st_size, cls._md5(path))

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.name}, {self.size}, ' \
               f'{self.md5})'


class Directory(Entity):
    """
    Represents a directory.
    """

    def __init__(self, name, children: Optional[Dict[str, Entity]] = None):
        """
        Initialise a new directory entity.

        :param name: The name of the directory, e.g. "photos".
        :param children: Entities within this directory. Keys are entity names,
                         whose values correspond to the entity object. Defaults
                         to no children, i.e. an empty directory.
        """
        super().__init__(name)
        self.children = children or {}  # type: Dict[str, Entity]

    def walk_paths(self,
                   base: Optional[pathlib.PurePath] = pathlib.PurePath()) \
            -> Iterator[pathlib.PurePath]:
        yield from itertools.chain.from_iterable(
            [child._walk_paths(base) for child in self.children.values()])

    def _walk_paths(self, base: pathlib.PurePath) \
            -> Iterator[pathlib.PurePath]:
        us = base / self.name
        yield us
        yield from itertools.chain.from_iterable(
            [child.walk_paths(us) for child in self.children.values()])

    def walk_files(self) -> Iterator[File]:
        for child in self.children.values():
            yield from child.walk_files()

    @classmethod
    def from_path(cls, path: pathlib.Path) -> 'Directory':
        """
        Create a directory entity from a directory path.

        :param path: The path of the directory.
        :return: A directory entity instance representing the directory.
        :raises ValueError: If the path does not point to a directory.
        """
        if not path.is_dir():
            raise ValueError('Path does not point to a directory')
        return Directory(path.name, {entity.name: Entity.from_path(entity)
                                     for entity in path.iterdir()})

    def print_hierarchy(self, level: int = 0, file: IO[str] = sys.stdout) \
            -> None:
        print(' ' * self._INDENT_SIZE * level + str(self), file=file)
        for name, child in self.children.items():
            child.print_hierarchy(level + 1, file)


class Root(Directory):
    """
    Represents a nameless directory at the root of a hierarchy. This is useful
    as a substitute for a list of entities representing the contents of the
    root directory.
    """

    def __init__(self, children: Dict[str, Entity]):
        """
        Initialise a root directory.

        :param children: Child entities within this directory. Names correspond
                         to entity names; their values are the entity objects
                         themselves.
        """
        # children is required as an empty root directory isn't very useful
        super().__init__('', children)

    @classmethod
    def from_path(cls, path: pathlib.Path) -> 'Root':
        """
        Create a root directory entity from a directory path.

        :param path: The path of the directory.
        :return: A root directory entity instance representing the directory.
        :raises ValueError: If the path does not point to a directory.
        """
        if not path.is_dir():
            raise ValueError('Path does not point to a directory')
        return Root({entity.name: Entity.from_path(entity)
                     for entity in path.iterdir()})



# -*- coding: utf-8 -*-
from typing import TypeVar, Generic
import abc

from wood.comparison import Comparison
from wood.entities import Entity

L = TypeVar('L', bound=Entity)
R = TypeVar('R', bound=Entity)


class Syncer(Generic[L, R], metaclass=abc.ABCMeta):
    """
    Implemented by things that know how to enact the changes in a comparison
    somewhere.
    """

    def sync(self, comparison: Comparison[L, R]) -> None:
        """
        Synchronise based on the changes in a comparison.

        :param comparison: The comparison to enact.
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


def _read_file(name, encoding='utf-8') -> str:
    """
    Read the contents of a file.

    :param name: The name of the file in the current directory.
    :param encoding: The encoding of the file; defaults to utf-8.
    :return: The contents of the file.
    """
    with open(name, encoding=encoding) as f:
        return f.read()


setup(
    name='wood',
    version='1.0.0-rc.1',
    description='Compare directories, efficiently sync changes to AWS, and '
                'invalidate CDNs.',
    long_description=_read_file('README.rst'),
    license='MIT',
    url='https://github.com/gebn/wood',
    author='George Brighton',
    author_email='oss@gebn.co.uk',
    packages=find_packages(),
    zip_safe=True,
    install_requires=[
        'backoff',
        'boto3',
        'botocore',
        'more-itertools',
        'requests'
    ],
    test_suite='nose.collector',
    tests_require=[
        'coverage',
        'coveralls',
        'mock',
        'nose'
    ],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



# -*- coding: utf-8 -*-
import logging
import itertools
from typing import List

import requests
import backoff

from wood import util
from wood.comparison import Comparison
from wood.entities import Entity
from wood.invalidate import Invalidator


logger = logging.getLogger(__name__)


class CloudflareInvalidator(Invalidator):
    """
    Invalidates URLs within a Cloudflare zone.
    """

    _API_BASE = 'https://api.cloudflare.com'
    _MAX_INVALIDATIONS_PER_REQUEST = 30

    def __init__(self, session: requests.Session, email: str, key: str,
                 zone: str, prefix: str):
        """
        Initialise a new Cloudflare cache invalidator.

        :param session: The requests session to use when interacting with
                        Cloudflare.
        :param zone: The zone ID to purge from.
        :param prefix: The full URL prefix to append asset paths to,
                       e.g. https://example.com/webroot/. Should always end
                       with a trailing slash.
        """
        self._session = session
        self._email = email
        self._key = key
        self._zone = zone
        self._prefix = prefix

    def invalidate(self, comparison: Comparison[Entity, Entity]) -> None:
        """
        Invalidate paths in a zone. See https://api.cloudflare.com
        /#zone-purge-individual-files-by-url-and-cache-tags

        :param comparison: The comparison whose changes to invalidate.
        :raises requests.exceptions.RequestException: On request failure.
        :raises RuntimeError: If the request succeeded but could not be carried
                              out.
        """

        @backoff.on_exception(backoff.expo,
                              requests.exceptions.RequestException,
                              max_tries=5,
                              giveup=lambda e:
                              400 <= e.response.status_code < 500)
        def _request(chunk: List[str]) -> requests.Response:
            """
            Send a purge cache request to Cloudflare. This method will
            automatically retry with a back-off in case of server-side error.

            :param chunk: The list of paths to purge. These should not have a
                          leading slash, and will be combined with the prefix
                          to form a URL.
            :return: Cloudflare's response to our successful request.
            :raises requests.exceptions.RequestException: If the request fails
                                                          on the 5th attempt.
            """
            response = self._session.delete(
                f'{self._API_BASE}/client/v4/zones/{self._zone}/purge_cache',
                headers={
                    'X-Auth-Email': self._email,
                    'X-Auth-Key': self._key
                },
                json={
                    'files': [self._prefix + path for path in chunk]
                })
            response.raise_for_status()
            return response

        paths = itertools.chain(comparison.deleted(), comparison.modified())
        for chunk_ in util.chunk(paths, self._MAX_INVALIDATIONS_PER_REQUEST):
            chunk_ = list(chunk_)
            if not chunk_:
                # nothing to do
                return
            logger.info('Invalidating %d paths (%s)', len(chunk_),
                        ', '.join(chunk_))
            response_ = _request(chunk_)
            logger.debug('Cloudflare invalidation response [%d]: %s',
                         response_.status_code,
                         response_.text)
            json_ = response_.json()
            if not json_['success']:
                # this would be strange - the API returned a success response
                # code, but success was not "true"
                # TODO more appropriate exception, with handling upstream
                raise RuntimeError('Cloudflare reported failure')
            logger.info('Created invalidation %s', json_['result']['id'])



# -*- coding: utf-8 -*-
import os

import pathlib
from typing import Union

from pkg_resources import get_distribution, DistributionNotFound

from wood.entities import Root as _Root, Entity as _Entity

from wood.comparison import Comparison
from wood.integrations import cloudflare, cloudfront, s3
root = _Root.from_path
entity = _Entity.from_path

__title__ = 'wood'
__author__ = 'George Brighton'
__license__ = 'MIT'
__copyright__ = 'Copyright 2017 George Brighton'


# adapted from http://stackoverflow.com/a/17638236
try:
    dist = get_distribution(__title__)
    path = os.path.normcase(dist.location)
    pwd = os.path.normcase(__file__)
    if not pwd.startswith(os.path.join(path, __title__)):
        raise DistributionNotFound()
    __version__ = dist.version
except DistributionNotFound:
    __version__ = 'unknown'


def compare(left: Union[str, pathlib.Path, _Entity],
            right: Union[str, pathlib.Path, _Entity]) -> Comparison:
    """
    Compare two paths.

    :param left: The left side or "before" entity.
    :param right: The right side or "after" entity.
    :return: A comparison details what has changed from the left side to the
             right side.
    """

    def normalise(param: Union[str, pathlib.Path, _Entity]) -> _Entity:
        """
        Turns any one of a number of types of input into an entity.

        :param param: The input - either a path string, a path object, or a
                      full blown entity.
        :return: The input param as an entity.
        """
        if isinstance(param, str):
            param = pathlib.Path(param)
        if isinstance(param, pathlib.Path):
            param = entity(param)
        return param

    return Comparison.compare(normalise(left), normalise(right))
